@@ -423,11 +423,19 @@ MODULE_SCOPE int tclpkcs11_close_session(struct tclpkcs11_handle *handle) {
 MODULE_SCOPE void *tclpkcs11_int_load_module(const char *pathname) {
 #if defined(TCL_INCLUDES_LOADFILE)
 	int tcl_rv;
+	Tcl_Obj *pathnameObj;
 	Tcl_LoadHandle *new_handle;
 
 	new_handle = (Tcl_LoadHandle *) ckalloc(sizeof(*new_handle));
 
-	tcl_rv = Tcl_LoadFile(NULL, Tcl_NewStringObj(pathname, -1), NULL, 0, NULL, new_handle);
+	pathnameObj = Tcl_NewStringObj(pathname, -1);
+
+	Tcl_IncrRefCount(pathnameObj);
+
+	tcl_rv = Tcl_LoadFile(NULL, pathnameObj, NULL, 0, NULL, new_handle);
+
+	Tcl_DecrRefCount(pathnameObj);
+
 	if (tcl_rv != TCL_OK) {
 		return(NULL);
 	}
@@ -535,6 +543,8 @@ MODULE_SCOPE int tclpkcs11_load_module(ClientData cd, Tcl_Interp *interp, int ob
 	if (!getFuncList) {
 		Tcl_SetObjResult(interp, Tcl_NewStringObj("unable to locate C_GetFunctionList symbol in PKCS#11 module", -1));
 
+		tclpkcs11_int_unload_module(handle);
+
 		return(TCL_ERROR);
 	}
 
@@ -542,18 +552,24 @@ MODULE_SCOPE int tclpkcs11_load_module(ClientData cd, Tcl_Interp *interp, int ob
 	if (chk_rv != CKR_OK) {
 		Tcl_SetObjResult(interp, tclpkcs11_pkcs11_error(chk_rv));
 
+		tclpkcs11_int_unload_module(handle);
+
 		return(TCL_ERROR);
 	}
 
 	if (!pkcs11_function_list) {
 		Tcl_SetObjResult(interp, Tcl_NewStringObj("C_GetFunctionList returned invalid data", -1));
 
+		tclpkcs11_int_unload_module(handle);
+
 		return(TCL_ERROR);
 	}
 
 	if (!pkcs11_function_list->C_Initialize) {
 		Tcl_SetObjResult(interp, Tcl_NewStringObj("C_GetFunctionList returned incomplete data", -1));
 
+		tclpkcs11_int_unload_module(handle);
+
 		return(TCL_ERROR);
 	}
 
@@ -569,6 +585,8 @@ MODULE_SCOPE int tclpkcs11_load_module(ClientData cd, Tcl_Interp *interp, int ob
 	if (chk_rv != CKR_OK) {
 		Tcl_SetObjResult(interp, tclpkcs11_pkcs11_error(chk_rv));
 
+		tclpkcs11_int_unload_module(handle);
+
 		return(TCL_ERROR);
 	}
 
@@ -582,6 +600,12 @@ MODULE_SCOPE int tclpkcs11_load_module(ClientData cd, Tcl_Interp *interp, int ob
 	if (!tcl_handle_entry) {
 		Tcl_SetObjResult(interp, Tcl_NewStringObj("unable to create new hash entry", -1));
 
+		if (pkcs11_function_list->C_Finalize) {
+			pkcs11_function_list->C_Finalize(NULL);
+		}
+
+		tclpkcs11_int_unload_module(handle);
+
 		return(TCL_ERROR);
 	}
 
@@ -1552,6 +1576,89 @@ MODULE_SCOPE int tclpkcs11_decrypt(ClientData cd, Tcl_Interp *interp, int objc,
 	return(tclpkcs11_perform_pki(0, cd, interp, objc, objv));
 }
 
+MODULE_SCOPE int tclpkcs11_setpin(ClientData cd, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]) {
+	struct tclpkcs11_interpdata *interpdata;
+	struct tclpkcs11_handle *handle;
+	Tcl_HashEntry *tcl_handle_entry;
+	Tcl_Obj *tcl_handle, *tcl_slotid, *tcl_oldpassword, *tcl_newpassword;
+	long slotid_long;
+	char *oldpassword, *newpassword;
+	int oldpassword_len, newpassword_len;
+	int tcl_rv;
+
+	CK_SLOT_ID slotid;
+	CK_RV chk_rv;
+
+	if (!cd) {
+		Tcl_SetObjResult(interp, Tcl_NewStringObj("invalid clientdata", -1));
+
+		return(TCL_ERROR);
+	}
+
+	if (objc != 5) {
+		Tcl_SetObjResult(interp, Tcl_NewStringObj("wrong # args: should be \"pki::pkcs11::setpin handle slot oldpassword newpassword\"", -1));
+
+		return(TCL_ERROR);
+	}
+
+	tcl_handle = objv[1];
+	tcl_slotid = objv[2];
+	tcl_oldpassword = objv[3];
+	tcl_newpassword = objv[4];
+
+	interpdata = (struct tclpkcs11_interpdata *) cd;
+
+	tcl_handle_entry = Tcl_FindHashEntry(&interpdata->handles, (const char *) tcl_handle);
+	if (!tcl_handle_entry) {
+		Tcl_SetObjResult(interp, Tcl_NewStringObj("invalid handle", -1));
+
+		return(TCL_ERROR);
+	}
+
+	handle = (struct tclpkcs11_handle *) Tcl_GetHashValue(tcl_handle_entry);
+	if (!handle) {
+		Tcl_SetObjResult(interp, Tcl_NewStringObj("invalid handle", -1));
+
+		return(TCL_ERROR);
+	}
+
+	tcl_rv = Tcl_GetLongFromObj(interp, tcl_slotid, &slotid_long);
+	if (tcl_rv != TCL_OK) {
+		return(tcl_rv);
+	}
+
+	slotid = slotid_long;
+
+	chk_rv = tclpkcs11_start_session(handle, slotid);
+	if (chk_rv != CKR_OK) {
+		Tcl_SetObjResult(interp, tclpkcs11_pkcs11_error(chk_rv));
+
+		return(TCL_ERROR);
+	}
+
+	oldpassword = Tcl_GetStringFromObj(tcl_oldpassword, &oldpassword_len);
+	newpassword = Tcl_GetStringFromObj(tcl_newpassword, &newpassword_len);
+
+	chk_rv = handle->pkcs11->C_SetPIN(handle->session, (CK_UTF8CHAR_PTR) oldpassword, oldpassword_len, (CK_UTF8CHAR_PTR) newpassword, newpassword_len);
+	switch (chk_rv) {
+		case CKR_OK:
+		case CKR_USER_ALREADY_LOGGED_IN:
+			Tcl_SetObjResult(interp, Tcl_NewBooleanObj(1));
+
+			break;
+		case CKR_PIN_INCORRECT:
+			Tcl_SetObjResult(interp, Tcl_NewBooleanObj(0));
+
+			break;
+		default:
+			Tcl_SetObjResult(interp, tclpkcs11_pkcs11_error(chk_rv));
+
+			return(TCL_ERROR);
+	}
+
+	return(TCL_OK);
+}
+
 MODULE_SCOPE void tclpkcs11_unloadall(ClientData cd) {
 	struct tclpkcs11_interpdata *interpdata;
 	struct tclpkcs11_handle *handle;
@@ -1658,6 +1765,11 @@ int Tclpkcs11_Init(Tcl_Interp *interp) {
 		return(TCL_ERROR);
 	}
 
+	tclCreatComm_ret = Tcl_CreateObjCommand(interp, "pki::pkcs11::setpin", tclpkcs11_setpin, interpdata, NULL);
+	if (!tclCreatComm_ret) {
+		return(TCL_ERROR);
+	}
+
 	/* Create an exit handler to unload and close all PKCS#11 modules */
 	Tcl_CreateExitHandler(tclpkcs11_unloadall, interpdata);
 
